#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['deal',
 'deal._cli',
 'deal._runtime',
 'deal.introspection',
 'deal.linter',
 'deal.linter._extractors']

package_data = \
{'': ['*'],
 'deal.introspection': ['.hypothesis/*',
                        '.hypothesis/unicode_data/*',
                        '.hypothesis/unicode_data/13.0.0/*',
                        '.mypy_cache/*',
                        '.mypy_cache/3.9/*',
                        '.mypy_cache/3.9/_pytest/*',
                        '.mypy_cache/3.9/_pytest/_code/*',
                        '.mypy_cache/3.9/_pytest/_io/*',
                        '.mypy_cache/3.9/_pytest/assertion/*',
                        '.mypy_cache/3.9/_pytest/config/*',
                        '.mypy_cache/3.9/_pytest/mark/*',
                        '.mypy_cache/3.9/_typeshed/*',
                        '.mypy_cache/3.9/asyncio/*',
                        '.mypy_cache/3.9/attr/*',
                        '.mypy_cache/3.9/collections/*',
                        '.mypy_cache/3.9/concurrent/*',
                        '.mypy_cache/3.9/concurrent/futures/*',
                        '.mypy_cache/3.9/ctypes/*',
                        '.mypy_cache/3.9/deal/*',
                        '.mypy_cache/3.9/deal/_runtime/*',
                        '.mypy_cache/3.9/deal/introspection/*',
                        '.mypy_cache/3.9/distutils/*',
                        '.mypy_cache/3.9/docutils/*',
                        '.mypy_cache/3.9/docutils/languages/*',
                        '.mypy_cache/3.9/docutils/parsers/*',
                        '.mypy_cache/3.9/docutils/parsers/rst/*',
                        '.mypy_cache/3.9/docutils/readers/*',
                        '.mypy_cache/3.9/docutils/transforms/*',
                        '.mypy_cache/3.9/docutils/utils/*',
                        '.mypy_cache/3.9/docutils/writers/*',
                        '.mypy_cache/3.9/email/*',
                        '.mypy_cache/3.9/html/*',
                        '.mypy_cache/3.9/http/*',
                        '.mypy_cache/3.9/hypothesis/*',
                        '.mypy_cache/3.9/hypothesis/extra/*',
                        '.mypy_cache/3.9/hypothesis/extra/django/*',
                        '.mypy_cache/3.9/hypothesis/internal/*',
                        '.mypy_cache/3.9/hypothesis/internal/conjecture/*',
                        '.mypy_cache/3.9/hypothesis/internal/conjecture/dfa/*',
                        '.mypy_cache/3.9/hypothesis/internal/conjecture/shrinking/*',
                        '.mypy_cache/3.9/hypothesis/strategies/*',
                        '.mypy_cache/3.9/hypothesis/strategies/_internal/*',
                        '.mypy_cache/3.9/hypothesis/utils/*',
                        '.mypy_cache/3.9/hypothesis/vendor/*',
                        '.mypy_cache/3.9/importlib/*',
                        '.mypy_cache/3.9/importlib_metadata/*',
                        '.mypy_cache/3.9/importlib_resources/*',
                        '.mypy_cache/3.9/iniconfig/*',
                        '.mypy_cache/3.9/jinja2/*',
                        '.mypy_cache/3.9/json/*',
                        '.mypy_cache/3.9/logging/*',
                        '.mypy_cache/3.9/markupsafe/*',
                        '.mypy_cache/3.9/multiprocessing/*',
                        '.mypy_cache/3.9/numpy/*',
                        '.mypy_cache/3.9/numpy/compat/*',
                        '.mypy_cache/3.9/numpy/core/*',
                        '.mypy_cache/3.9/numpy/distutils/*',
                        '.mypy_cache/3.9/numpy/fft/*',
                        '.mypy_cache/3.9/numpy/lib/*',
                        '.mypy_cache/3.9/numpy/linalg/*',
                        '.mypy_cache/3.9/numpy/ma/*',
                        '.mypy_cache/3.9/numpy/matrixlib/*',
                        '.mypy_cache/3.9/numpy/polynomial/*',
                        '.mypy_cache/3.9/numpy/random/*',
                        '.mypy_cache/3.9/numpy/testing/*',
                        '.mypy_cache/3.9/numpy/typing/*',
                        '.mypy_cache/3.9/os/*',
                        '.mypy_cache/3.9/packaging/*',
                        '.mypy_cache/3.9/py/*',
                        '.mypy_cache/3.9/pytest/*',
                        '.mypy_cache/3.9/sphinx/*',
                        '.mypy_cache/3.9/sphinx/builders/*',
                        '.mypy_cache/3.9/sphinx/directives/*',
                        '.mypy_cache/3.9/sphinx/domains/*',
                        '.mypy_cache/3.9/sphinx/environment/*',
                        '.mypy_cache/3.9/sphinx/environment/adapters/*',
                        '.mypy_cache/3.9/sphinx/environment/collectors/*',
                        '.mypy_cache/3.9/sphinx/ext/*',
                        '.mypy_cache/3.9/sphinx/ext/autodoc/*',
                        '.mypy_cache/3.9/sphinx/locale/*',
                        '.mypy_cache/3.9/sphinx/pycode/*',
                        '.mypy_cache/3.9/sphinx/search/*',
                        '.mypy_cache/3.9/sphinx/testing/*',
                        '.mypy_cache/3.9/sphinx/transforms/*',
                        '.mypy_cache/3.9/sphinx/util/*',
                        '.mypy_cache/3.9/sphinx/util/stemmer/*',
                        '.mypy_cache/3.9/toml/*',
                        '.mypy_cache/3.9/typeguard/*',
                        '.mypy_cache/3.9/unittest/*',
                        '.mypy_cache/3.9/urllib/*',
                        '.mypy_cache/3.9/zoneinfo/*'],
 'deal.linter': ['stubs/*',
                 'stubs/aiohttp/*',
                 'stubs/cpython/*',
                 'stubs/cpython/asyncio/*',
                 'stubs/cpython/http/*',
                 'stubs/cpython/sqlite3/*',
                 'stubs/cpython/urllib/*',
                 'stubs/dateutil/*',
                 'stubs/dateutil/parser/*',
                 'stubs/dateutil/tz/*',
                 'stubs/dateutil/zoneinfo/*',
                 'stubs/marshmallow/*',
                 'stubs/pytz/*',
                 'stubs/requests/*',
                 'stubs/urllib3/*',
                 'stubs/urllib3/contrib/*',
                 'stubs/urllib3/util/*']}

install_requires = \
['astroid', 'deal-solver', 'hypothesis', 'pygments', 'typeguard', 'vaa>=0.2.1']

extras_require = \
{'docs': ['m2r2', 'myst-parser', 'sphinx==3.5.*', 'sphinx-rtd-theme==0.5.*'],
 'lint': ['flake8',
          'flake8-commas',
          'flake8-quotes',
          'mypy>=0.900',
          'mypy_test>=0.1.1',
          'isort'],
 'test': ['coverage[toml]',
          'pytest-cov',
          'isort',
          'marshmallow',
          'pytest',
          'urllib3',
          'sphinx']}

entry_points = \
{'flake8.extension': ['DEAL = deal.linter:Checker']}

setup(name='deal',
      version='4.19.2',
      description='**Deal** is a Python library for [design by contract][wiki] (DbC) programming.',
      author='Gram',
      author_email='gram@orsinium.dev',
      url='https://github.com/life4/deal',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.7',
     )
